import pytest

from src.pg_stage.obfuscator import Obfuscator


@pytest.fixture(autouse=True)
def obfuscator_object_with_delete_tables_by_pattern():
    return Obfuscator(delete_tables_by_pattern=[r'_table', r'schema\.\_table'])


@pytest.mark.parametrize("table_name", ['table', 'schema.table'])
def test_parse_copy_values_with_delete_comment(obfuscator_object: Obfuscator, table_name: str):
    """
    Arrange: Строка копирования данных в таблицу, которую необходимо удалить
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Флаг _is_delete активен
    """
    obfuscator_object._delete_tables.add(table_name)
    obfuscator_object._parse_line(
        line=f'COPY {table_name} (column_1, column_2) FROM stdin;',
    )
    assert obfuscator_object._is_delete  # nosec
    assert obfuscator_object._is_data  # nosec


@pytest.mark.parametrize("table_name", ['table', 'schema.table'])
def test_parse_copy_values_without_delete_comment(obfuscator_object: Obfuscator, table_name: str):
    """
    Arrange: Строка копирования данных в таблицу, которую необходимо удалить
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Флаг _is_delete активен
    """
    obfuscator_object._parse_line(
        line=f'COPY {table_name} (column_1, column_2) FROM stdin;',
    )
    assert not obfuscator_object._is_delete  # nosec
    assert obfuscator_object._is_data  # nosec


@pytest.mark.parametrize("table_name", ['_table', 'schema._table'])
def test_parse_copy_values_without_delete_tables_by_pattern(
    obfuscator_object_with_delete_tables_by_pattern: Obfuscator,
    table_name: str,
):
    """
    Arrange: Строка копирования данных в таблицу, которую необходимо удалить исходя из параметра
    delete_tables_by_pattern
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Флаг _is_delete активен
    """
    obfuscator_object_with_delete_tables_by_pattern._parse_line(
        line=f'COPY {table_name} (column_1, column_2) FROM stdin;',
    )
    assert obfuscator_object_with_delete_tables_by_pattern._is_delete  # nosec
    assert obfuscator_object_with_delete_tables_by_pattern._is_data  # nosec


@pytest.mark.parametrize("table_name", ['table', 'schema.table'])
def test_parse_copy_values_without_delete_tables_by_pattern_and_not_found_table(
    obfuscator_object_with_delete_tables_by_pattern: Obfuscator,
    table_name: str,
):
    """
    Arrange: Строка копирования данных в таблицу, которая не должна быть удалена, т.к. не входит в список
    delete_tables_by_pattern
    delete_tables_by_pattern
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Флаг _is_delete не активен
    """
    obfuscator_object_with_delete_tables_by_pattern._parse_line(
        line=f'COPY {table_name} (column_1, column_2) FROM stdin;',
    )
    assert not obfuscator_object_with_delete_tables_by_pattern._is_delete  # nosec
    assert obfuscator_object_with_delete_tables_by_pattern._is_data  # nosec



import uuid
from datetime import date

from src.pg_stage.obfuscator import Obfuscator


def test_parse_copy_values_with_delete_table(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблиц, данные которых необходимо удалить
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout нет данных из таблицы
    """
    with open('tests/sql/test_parse_copy_values_with_delete_tables.sql') as file:
        dump_sql = file.read()

    result = []
    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            result.append(new_line)

    assert_result = [
        'COMMENT ON TABLE table_1 IS \'anon: {"mutation_name": "delete"}\';',
        'COPY table_1 (id, message, recipient, notes) FROM stdin;',
        '\\.',
        '',
        'COMMENT ON TABLE schema.table_2 IS \'anon: {"mutation_name": "delete"}\';',
        'COPY schema.table_2 (id, message, recipient, notes) FROM stdin;',
        '\\.',
    ]
    assert assert_result == result  # nosec


def test_parse_copy_values_with_mutation(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблиц, в которых необходимо мутировать одно или несколько полей
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout данные полей мутированы
    """
    with open('tests/sql/test_parse_copy_values_with_mutation.sql') as file:
        dump_sql = file.read()

    emails = {'cj@example.com', 'leo@example.com', 'donna@example.com', 'charlie@example.com', 'fun@example.com'}
    birthdays = {'1996-10-02', '1996-10-03', '2005-01-02', '2005-10-02', '2022-12-31'}
    tokens = {
        '86a97ff982e87ed5af7d90ab2ce31d4e89a3af3e6a0490b067bb8213aea7a4ee0eeafae1d8fe3c6f990aead095092fcf852004b18'
        'e484ef22569aebf64c3747f',
        '630aaf10f3edcb396181535476bccfd369e1ee9281cbc218d0861e6117b9839219cb403abed3b59468b2564c229d78b0cb610d0e1'
        '6fa4a03dc4d93ed3d55deb3',
        '06fa5d0f3a4d62ff60717ff3315301129a5bcd152f63f073b67d2e6a835b098524b1d097367283c97ca75db9154cf61d6a918248e'
        'b68eaf8cd83d1a87cdd92dc',
        '02cbd8d1e9690c7786375f9f6ec4da6fa3c2ead70d41a7aa1d7fe631477b42635210f0e8fa7579f6f4b5afdb364d9465623a2861e'
        'cff75fa3a052c9b2a9fc2dd',
        'b09909b6e83938dd41ffb5e931eeb3d646b1856dfb74a81acb1697b6d8466468047fe92286e011a4634c71b8d8775c7d5a31e19ce'
        '111bd31d0a61a4faf93d6af',
    }
    crypted_password = (  # nosec
        '400$8$4c$c11df6facaefc6bc$93a657fb3c6e4cd1fd3125d3bd1ett18ffc4a8092e2616b9e5ca7954b2c52504'
    )
    password_salts = {
        'v8BMktHnOeokEBTy6As',
        'c5d6v3NS97D3wYkUltFQ',
        '2IcBeSh6IVVCJyZpqBt',
        'QOqWQ24CMiRNuPUz8f5',
        'uMYA6c3A7uGoI0aEbJ8R',
    }

    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line and 'COPY' not in new_line and 'COMMENT' not in new_line and '\\.' not in line:
            assert not any([email in new_line for email in emails])  # nosec
            assert not any([birthday in new_line for birthday in birthdays])  # nosec
            assert not any([token in new_line for token in tokens])  # nosec
            assert not any([password_salt in new_line for password_salt in password_salts])  # nosec
            assert crypted_password in new_line  # nosec


def test_parse_copy_values_with_relations(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблиц, в которых необходимо мутировать одно связанное поле
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout данные поле мутировано во всех таблицах одинаково и не задета таблица с похожими данными
    """
    with open('tests/sql/test_parse_copy_values_with_relations.sql') as file:
        dump_sql = file.read()

    result = []
    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            result.append(new_line)

    index_copy_table4 = result.index('20f654fe-b27d-4051-9fd4-000000000001\t111n\tLourense') - 1
    assert 'table_4' in result[index_copy_table4]  # nosec
    assert (  # nosec
        len([line for line in result if line == '20f654fe-b27d-4051-9fd4-000000000001\t111n\tLourense']) == 1
    )
    assert len([line for line in result if line == '20f654fe-b27d-4051-9fd4-000000000002\t222n\tKent']) == 1  # nosec
    assert (  # nosec
        len({line for line in result if '20f654fe-b27d-4051-9fd4-000000000001' in line and '111n' not in line}) == 1
    )
    assert (  # nosec
        len({line for line in result if '20f654fe-b27d-4051-9fd4-000000000002' in line and '222n' not in line}) == 1
    )


def test_parse_copy_values_with_self_relation(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблицы, в которой необходимо мутировать данные с связанным полем внутри таблицы
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout данные поле мутировано в двух столбцах одинаково
    """
    with open('tests/sql/test_parse_copy_values_with_self_relation.sql') as file:
        dump_sql = file.read()

    result = []
    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            result.append(new_line)

    prepared_result = [
        line.split('\t')[1:]
        for line in result
        if '20f654fe-b27d-4051-9fd4-000000000001' in line or '20f654fe-b27d-4051-9fd4-000000000002' in line
    ]
    assert prepared_result[0][0] == prepared_result[0][1]  # nosec
    assert prepared_result[1][0] == prepared_result[1][1]  # nosec


def test_two_mutation_for_one_column(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблицы, в которой необходимо мутировать данные с двумя видами мутации
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout данные поля мутированы
    """
    with open('tests/sql/test_two_mutation_for_one_column.sql') as file:
        dump_sql = file.read()

    result = []
    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            result.append(new_line)

    prepared_result = [
        line.split('\t')[1:]
        for line in result
        if '20f654fe-b27d-4051-9fd4-000000000001' in line
        or '20f654fe-b27d-4051-9fd4-000000000002' in line
        or '20f654fe-b27d-4051-9fd4-000000000003' in line
    ]

    assert prepared_result[0][0] != '79999999999'  # nosec
    assert prepared_result[1][0] != 'test@mail.ru'  # nosec
    assert prepared_result[2][0] != '89999999999'  # nosec


def test_obfuscate_uuid_by_source_column(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблицы, в которой определены мутации uuid на основе мутации другой колонки
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout данные поля мутированы на основе мутированного значения другой колонки
    """
    with open('tests/sql/test_obfuscate_uuid_by_source_value.sql') as file:
        dump_sql = file.read()

    result = []
    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            result.append(new_line)

    prepared_result = [line.split('\t') for line in result if '111n' in line or '222n' in line or '333n' in line]

    date_today = date.today()
    uuid_namespace = uuid.UUID(str('6ba7b810-9dad-11d1-80b4-00c04fd430c8'))
    for column_uuid, phone, last_name in prepared_result:
        assert column_uuid == str(uuid.uuid5(uuid_namespace, f'{phone}-{date_today}'))  # nosec



import json
import re
import sys
from collections import defaultdict
from typing import Dict, List, Optional, Set
from uuid import uuid4

from pg_stage.mutator import Mutator
from pg_stage.typing import ConditionTypeMany, MapTablesValueTypeMany


class Obfuscator:
    """Главный класс для работы с обфускатором."""

    copy_parse_pattern = r'COPY ([\d\w\_\.]+) \(([\w\W]+)\) FROM stdin;'
    comment_table_parse_pattern = r'COMMENT ON TABLE ([\d\w\_\.]*) IS \'anon: ([\w\W]*)\'\;'
    comment_column_parse_pattern = r'COMMENT ON COLUMN ([\d\w\_\.]+) IS \'anon: ([\w\W]*)\'\;'

    def __init__(
        self,
        delimiter: str = '\t',
        locale: str = 'en',
        delete_tables_by_pattern: Optional[List[str]] = None,
    ) -> None:
        """
        Метод инициализации класса.
        :param delimiter: разделитель
        :param locale: локализация для Faker
        :param delete_tables_by_pattern: список таблиц, которые нужно очистить по паттерну
        """
        self.delimiter = delimiter
        self.delete_tables_by_pattern: List[str] = delete_tables_by_pattern or []
        self._map_tables: Dict[str, Dict[str, MapTablesValueTypeMany]] = defaultdict(dict)
        self._mutator = Mutator(locale=locale)
        self._relation_values: Dict[str, str] = {}
        self._relation_fk: Dict[str, Dict[str, Dict[str, str]]] = defaultdict(dict)
        self._is_data: bool = False
        self._schema_name: Optional[str] = None
        self._table_name: str = ''
        self._table_columns: List[str] = []
        self._enumerate_table_columns: Dict[str, int] = {}
        self._delete_tables: Set[str] = set()
        self._is_delete: bool = False

    def _prepare_variables(self, *, line: str) -> Optional[str]:
        """
        Метод для установки начальных значений основных переменных.
        :param line: строка sql
        :return: строка sql
        """
        self._is_data = False
        self._table_name = ''
        self._table_columns = []
        self._enumerate_table_columns = {}
        self._is_delete = False
        return line

    def _checking_conditions(self, *, conditions: ConditionTypeMany, table_values: List[str]) -> bool:
        """
        Метод для проверки условий обфускации.
        :param conditions: условия
        :param table_values: значения вставляемой строки из дампа
        :return: флаг об выполнение обфускации для столбца
        """
        if not conditions:
            return True

        flag = False
        for condition in conditions:
            if flag:
                break

            column_name = condition['column_name']
            operation = condition['operation']
            value = condition['value']

            column_value = table_values[self._enumerate_table_columns[column_name]]
            if operation == 'equal':
                flag = column_value == value
                continue

            if operation == 'not_equal':
                flag = column_value != value
                continue

            if operation == 'by_pattern':
                flag = re.search(pattern=value, string=column_value) is not None
                continue

            raise ValueError('Invalid condition operation')

        return flag

    def _parse_comment_column(self, *, line: str) -> str:
        """
        Метод для обработки комментария колонки для составления карты.
        :param line: строка sql
        :return: строка sql
        """
        result = re.search(pattern=self.comment_column_parse_pattern, string=line)
        if not result:
            return line

        try:
            mutations_params = json.loads(result.group(2))
        except ValueError:
            return line

        for mutation_param in mutations_params:
            mutation_name = mutation_param['mutation_name']
            mutation_func = getattr(self._mutator, f'mutation_{mutation_name}', None)
            if not mutation_func:
                raise ValueError(f'Not found mutation {mutation_name}.')

            try:
                table_name, column_name = result.group(1).split('.')
            except ValueError:
                schema_name, table_name, column_name = result.group(1).split('.')
                table_name = f'{schema_name}.{table_name}'

            self._map_tables[table_name].setdefault(column_name, [])
            self._map_tables[table_name][column_name].append(
                {
                    'mutation_name': mutation_name,
                    'mutation_func': mutation_func,
                    'mutation_kwargs': mutation_param.get('mutation_kwargs', {}),
                    'mutation_relations': mutation_param.get('relations', []),
                    'mutation_conditions': mutation_param.get('conditions', []),
                },
            )

        return line

    def _parse_comment_table(self, *, line: str) -> str:
        """
        Метод для обработки комментария таблицы.
        :param line: строка sql
        :return: строка sql
        """
        result = re.search(pattern=self.comment_table_parse_pattern, string=line)
        if not result:
            return line

        try:
            mutation_params = json.loads(result.group(2))
        except ValueError:
            return line

        mutation_name = mutation_params['mutation_name']
        if mutation_name == 'delete':
            self._delete_tables.add(result.group(1))

        return line

    def _sort_columns_by_source_column_exists(self, table_mutations_by_column: dict) -> list:
        """
        Метод для сортировки столбцов на основе наличия параметра `source_column` в аргументах мутации.
        :return: Список с верным порядком прохождения столбцов
        """
        independent_columns = []
        dependent_columns = []
        for column_name in self._enumerate_table_columns:
            mutations_for_column = table_mutations_by_column.get(column_name)
            if mutations_for_column and 'source_column' in mutations_for_column[0].get('mutation_kwargs', {}):
                dependent_columns.append(column_name)
                continue
            independent_columns.append(column_name)
        return independent_columns + dependent_columns

    def _prepared_data(self, *, line: str) -> Optional[str]:
        """
        Метод для обработки данных.
        :return: Новая строка с данными
        """
        if self._is_delete:
            return None

        table_mutations_by_column = self._map_tables.get(self._table_name)
        if not table_mutations_by_column:
            return line

        sorted_columns = self._sort_columns_by_source_column_exists(table_mutations_by_column)

        obfuscated_values = {}
        table_values = line.split(self.delimiter)
        for column_name in sorted_columns:
            column_index = self._enumerate_table_columns[column_name]
            mutations_for_column = table_mutations_by_column.get(column_name)
            if not mutations_for_column:
                obfuscated_values[column_name] = table_values[column_index]
                continue

            is_obfuscated = False
            len_mutations_for_column = len(mutations_for_column)
            for mutation_index, mutation_for_column in enumerate(mutations_for_column):
                if is_obfuscated:
                    break

                mutation_func = mutation_for_column['mutation_func']
                mutation_kwargs = mutation_for_column['mutation_kwargs']
                mutation_relations = mutation_for_column['mutation_relations']
                mutation_conditions = mutation_for_column['mutation_conditions']
                if not self._checking_conditions(conditions=mutation_conditions, table_values=table_values):
                    if mutation_index + 1 == len_mutations_for_column:
                        obfuscated_values[column_name] = table_values[column_index]
                        break

                    continue

                if 'source_column' in mutation_kwargs:
                    mutation_kwargs['obfuscated_values'] = obfuscated_values

                if not mutation_relations:
                    is_obfuscated = True
                    obfuscated_values[column_name] = mutation_func(**mutation_kwargs)
                    continue

                new_value = None
                for mutation_relation in mutation_relations:
                    key_table = f'{mutation_relation["table_name"]}:{mutation_relation["column_name"]}'
                    from_column_name = mutation_relation['from_column_name']
                    to_column_name = mutation_relation['to_column_name']
                    relation_key_value = table_values[self._enumerate_table_columns[from_column_name]]
                    relation_fk = self._relation_fk.get(key_table, {}).get(to_column_name, {}).get(relation_key_value)
                    if not relation_fk:
                        continue

                    new_value = self._relation_values.get(relation_fk)
                    if new_value is None:
                        raise ValueError('Invalid relation fk!')

                    break

                if new_value is None:
                    relation_fk = str(uuid4())
                    new_value = mutation_func(**mutation_kwargs)
                    for mutation_relation in mutation_relations:
                        key_table = f'{self._table_name}:{column_name}'
                        from_column_name = mutation_relation['from_column_name']
                        relation_key_value = table_values[self._enumerate_table_columns[from_column_name]]
                        self._relation_fk[key_table].setdefault(from_column_name, {})
                        self._relation_fk[key_table][from_column_name][relation_key_value] = relation_fk

                    self._relation_values[relation_fk] = new_value

                obfuscated_values[column_name] = new_value
                is_obfuscated = True

        sorted_result: list[str] = []
        for column_name, column_index in self._enumerate_table_columns.items():
            sorted_result.append(obfuscated_values[column_name])

        return self.delimiter.join(sorted_result)

    def _parse_copy_values(self, *, line: str) -> Optional[str]:
        """
        Метод для обработки строк с COPY.
        :param line: строка sql
        :return: строка sql
        """
        result = re.search(pattern=self.copy_parse_pattern, string=line)
        if not result:
            return None

        try:
            schema_name, _ = result.group(1).split('.')
        except ValueError:
            schema_name = None

        if self._schema_name != schema_name:
            # Если произошла смена схемы БД, то сбрасываем накопившиеся уникальные значения для ускорения работы
            self._mutator.clear_unique_values()

        self._schema_name = schema_name
        self._table_name = result.group(1)
        self._table_columns = [item.strip() for item in result.group(2).split(',')]
        self._enumerate_table_columns = {column_name: index for index, column_name in enumerate(self._table_columns)}
        self._is_delete = self._table_name in self._delete_tables or any(
            re.search(pattern, self._table_name) for pattern in self.delete_tables_by_pattern
        )
        self._is_data = True
        return line

    def _parse_line(self, *, line: str) -> Optional[str]:
        """
        Метод для парсинга строки из дампа.
        :param line: строка sql
        :return: обработанная строка sql
        """
        if line.startswith('\\.'):
            return self._prepare_variables(line=line)

        if self._is_data:
            return self._prepared_data(line=line)

        if line.startswith('COMMENT ON COLUMN'):
            return self._parse_comment_column(line=line)

        if line.startswith('COMMENT ON TABLE'):
            return self._parse_comment_table(line=line)

        if line.startswith('COPY'):
            return self._parse_copy_values(line=line)

        return line

    def run(self, *, stdin=None) -> None:
        """
        Метод для запуска обфускации.
        :param stdin: поток, с которого приходит информация в виде строк sql
        """
        if not stdin:
            stdin = sys.stdin

        for line in stdin:
            new_line = self._parse_line(line=line.rstrip('\n'))
            if isinstance(new_line, str):
                sys.stdout.write(new_line + '\n')



from src.pg_stage.obfuscator import Obfuscator


def test_equal_condition(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблиц, данные в которой необходимо обфусцировать
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout обфусцированы только те данные, условия которых выполняются
    """
    with open('tests/sql/test_equal_condition.sql') as file:
        dump_sql = file.read()

    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            if new_line.startswith('1'):
                assert 'test@mail.ru' not in new_line  # nosec

            if new_line.startswith('2'):
                assert 'test@mail.ru' in new_line  # nosec


def test_not_equal_condition(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблиц, данные в которой необходимо обфусцировать
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout обфусцированы только те данные, условия которых выполняются
    """
    with open('tests/sql/test_not_equal_condition.sql') as file:
        dump_sql = file.read()

    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            if new_line.startswith('1'):
                assert 'test@mail.ru' in new_line  # nosec

            if new_line.startswith('2'):
                assert 'test@mail.ru' not in new_line  # nosec


def test_by_pattern_condition(obfuscator_object: Obfuscator):
    """
    Arrange: Дамп таблиц, данные в которой необходимо обфусцировать
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: В stdout обфусцированы только те данные, условия которых выполняются
    """
    with open('tests/sql/test_by_pattern_condition.sql') as file:
        dump_sql = file.read()

    for line in dump_sql.splitlines():
        new_line = obfuscator_object._parse_line(line=line)
        if new_line is not None:
            if new_line.startswith('1'):
                assert 'test@mail.ru' not in new_line  # nosec

            if new_line.startswith('2'):
                assert 'test@mail.ru' not in new_line  # nosec



import pytest

from src.pg_stage.obfuscator import Obfuscator


@pytest.mark.parametrize("table_name, column_name", [('table', 'column'), ('schema.table', 'column')])
def test_parse_comment_column_with_mutation(
    obfuscator_object: Obfuscator,
    table_name: str,
    column_name: str,
):
    """
    Arrange: Строка, в которой есть комментарий к колонке
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Мутация добавлена в список _map_tables
    """
    obfuscator_object._parse_line(
        line=f'COMMENT ON COLUMN {table_name}.{column_name}' + ' IS \'anon: [{"mutation_name": "null"}]\';',
    )
    assert table_name in obfuscator_object._map_tables  # nosec
    assert column_name in obfuscator_object._map_tables[table_name]  # nosec
    assert 'null' == obfuscator_object._map_tables[table_name][column_name][0]['mutation_name']  # nosec


@pytest.mark.parametrize("table_name, column_name", [('table', 'column'), ('schema.table', 'column')])
def test_parse_comment_column_without_mutation(
    obfuscator_object: Obfuscator,
    table_name: str,
    column_name: str,
):
    """
    Arrange: Строка, в которой есть комментарий к колонке с неизвестной мутацией
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Вброшено исключение ValueError
    """
    with pytest.raises(ValueError):
        obfuscator_object._parse_line(
            line=f'COMMENT ON COLUMN {table_name}.{column_name}' + ' IS \'anon: [{"mutation_name": "not_found"}]\';',
        )


@pytest.mark.parametrize("table_name, column_name", [('table', 'column'), ('schema.table', 'column')])
def test_parse_comment_column_with_invalid_mutation(
    obfuscator_object: Obfuscator,
    table_name: str,
    column_name: str,
):
    """
    Arrange: Строка, в которой есть невалидный комментарий к колонке
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Мутация не добавлена в список _map_tables
    """
    obfuscator_object._parse_line(
        line=f'COMMENT ON COLUMN {table_name}.{column_name}' + ' IS \'anon: {\'mutation_name": "null"}\';',
    )
    assert table_name not in obfuscator_object._map_tables  # nosec



import datetime
import random
import uuid
from typing import Any, Callable, List, Optional

from mimesis import Address, Datetime, Internet, Numbers, Person
from mimesis.builtins import RussiaSpecProvider


class Mutator:
    """Класс с описанием основных методов для мутации значений полей."""

    min_value_smallint = -32768
    max_value_smallint = -32767
    min_value_integer = -2147483648
    max_value_integer = 2147483647
    min_value_bigint = -9223372036854775808
    max_value_bigint = 9223372036854775807
    min_value_smallserial = 1
    max_value_smallserial = 32767
    min_value_serial = 1
    max_value_serial = 2147483647
    min_value_bigserial = 1
    max_value_bigserial = 9223372036854775807

    def __init__(self, locale: str = 'en') -> None:
        """
        Метод инициализации класса.
        :param locale: локализация для Faker
        """
        self._locale = locale
        self._is_russian_locale = locale == 'ru'
        self._person = Person(locale=self._locale)
        self._address = Address(locale=self._locale)
        self._datetime = Datetime(locale=self._locale)
        self._internet = Internet()
        self._numeric = Numbers()
        self._russian_provider = RussiaSpecProvider()
        self._current_year = datetime.date.today().year
        self._now = datetime.datetime.now()
        self._today = self._now.date()
        self._cache = {}  # type: ignore
        self._unique_values = set()  # type: ignore

    def clear_unique_values(self) -> None:
        """Метод для сброса уникальных значений."""
        self._unique_values.clear()

    def _generate_unique_value(self, func: Callable[[Any], Any], *args: Any, **kwargs: Any) -> Any:
        """Метод для генерации уникального значения."""
        counter = 0
        while True:
            if counter > 1000:
                raise RecursionError('Exceeded the number of attempts to generate a unique value!')

            value = func(*args, **kwargs)
            if value not in self._unique_values:
                self._unique_values.add(value)
                break

            counter += 1

        return value

    @staticmethod
    def _random_int(a: int, b: int) -> int:
        b = b - a
        return int(random.random() * b) + a  # nosec

    def _generate_string_by_mask(self, mask: str, char: str = '@', digit: str = '#') -> str:
        """
        Метод для генерации строки по маске.
        :param mask: маска
        :param char: маска символов
        :param digit: маска цифр
        TODO: После обновления mimesis использовать mimesis.random.Random.generate_string_by_mask
        """
        char_code = ord(char)
        digit_code = ord(digit)

        if char_code == digit_code:
            raise ValueError('The same placeholder cannot be used for both numbers and characters.')

        _mask = mask.encode()
        code = bytearray(len(_mask))
        for i, p in enumerate(_mask):
            if p == char_code:
                a = self._random_int(65, 91)  # A-Z
            elif p == digit_code:
                a = self._random_int(48, 58)  # 0-9
            else:
                a = p
            code[i] = a
        return code.decode()

    def mutation_email(self, **kwargs: bool) -> str:
        """
        Метод для генерации email-а.
        :param kwargs:
            unique - сгенерировать уникальный email
        :return: email
        """
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._person.email)

        return self._person.email()

    @staticmethod
    def mutation_empty_string(**_: Any) -> str:
        """
        Метод для создания пустой строки.
        :param _: параметры генерации (не используются)
        :return: пустая строка
        """
        return ''

    @staticmethod
    def mutation_fixed_value(**kwargs: Any) -> str:
        """
        Метод для вставки значения из параметров.
        :param kwargs:
            value - значение, которое необходимо вернуть
        :return: строка со значением
        """
        return str(kwargs['value'])

    def mutation_full_name(self, **kwargs: bool) -> str:
        """
        Метод для генерации ФИО.
        :param kwargs:
            unique - сгенерировать уникальное ФИО
        :return: ФИО
        """
        if kwargs.get('unique'):
            while True:
                if self._is_russian_locale:
                    value = f'{self._person.full_name(reverse=True)} {self._russian_provider.patronymic()}'
                else:
                    value = self._person.full_name(reverse=True)

                if not set(value) & self._unique_values:
                    self._unique_values.add(value)
                    break

            return value

        if self._is_russian_locale:
            return f'{self._person.full_name(reverse=True)} {self._russian_provider.patronymic()}'

        return self._person.full_name(reverse=True)

    def mutation_first_name(self, **kwargs: bool) -> str:
        """
        Метод для генерации имени.
        :param kwargs:
            unique - сгенерировать уникальное имя
        :return: имя
        """
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._person.name)

        return self._person.name()

    def mutation_middle_name(self, **kwargs: bool) -> str:
        """
        Метод для генерации отчества (работает только с ru_RU).
        :param kwargs:
            unique - сгенерировать уникальное отчество
        :return: отчество
        """
        if not self._is_russian_locale:
            raise ValueError('Mutation middle_name dont work not Russian locale!')

        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._russian_provider.patronymic)

        return self._russian_provider.patronymic()

    def mutation_last_name(self, **kwargs: bool) -> str:
        """
        Метод для генерации фамилии.
        :param kwargs:
            unique - сгенерировать уникальную фамилию
        :return: фамилия
        """
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._person.surname)

        return self._person.surname()

    @staticmethod
    def mutation_null(**_: Any) -> str:
        """
        Метод для возвращения NULL значения.
        :param _: параметры генерации (не используются)
        :return: NULL
        """
        return '\\N'

    def mutation_phone_number(self, **kwargs: Any) -> str:
        """
        Метод для генерации номера телефона.
        :param kwargs:
            mask - формат номера, например +7 (XXX) XXX XX XX
            unique - сгенерировать уникальный номер
        :return: номер телефона
        """
        phone_format: str = kwargs['mask']
        unique: bool = kwargs.get('unique', False)
        if unique:
            return self._generate_unique_value(func=self._person.identifier, mask=phone_format)

        return self._person.identifier(mask=phone_format)

    def mutation_address(self, **kwargs: bool) -> str:
        """
        Метод для генерации адреса.
        :param kwargs:
            unique - сгенерировать уникальный адрес
        :return: адрес
        """
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._address.address)

        return self._address.address()

    def mutation_date(self, **kwargs: Any) -> str:
        """
        Метод для генерации даты между start и end годами.
        :param kwargs:
            start - год начала
            end - год окончания
            date_format - формат даты, по умолчанию '%Y-%m-%d'
        :return: дата в прошедшем времени
        """
        start: int = kwargs.get('start', self._now.year - 1)
        end: int = kwargs.get('end', self._now.year)
        date_format: int = kwargs.get('date_format', '%Y-%m-%d')
        unique: bool = kwargs.get('unique', False)
        if unique:
            return self._generate_unique_value(func=self._datetime.date, start=start, end=end).strftime(date_format)

        return self._datetime.date(start=start, end=end).strftime(date_format)

    def mutation_uri(self, **kwargs: Any) -> str:
        """
        Метод для генерации uri.
        :param kwargs:
            unique - сгенерировать уникальный uri
        :return: uri
        """
        max_length: int = kwargs.get('max_length', 2048)
        unique: bool = kwargs.get('unique', False)
        if unique:
            return self._generate_unique_value(func=self._internet.home_page)[:max_length]

        return self._internet.home_page()[:max_length]

    def mutation_ipv4(self, **kwargs: bool) -> str:
        """
        Метод для генерации ip-адреса 4 версии.
        :param kwargs:
            unique - сгенерировать уникальный ip-адрес
        :return: ip-адрес
        """
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._internet.ip_v4)

        return self._internet.ip_v4()

    def mutation_ipv6(self, **kwargs: bool) -> str:
        """
        Метод для формирования ip-адреса 6-й версии.
        :param kwargs:
            unique - сгенерировать уникальный ip-адрес
        :return: ip-адрес
        """
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._internet.ip_v6)

        return self._internet.ip_v6()

    @staticmethod
    def mutation_random_choice(**kwargs: List[Any]) -> str:
        """
        Метод для формирования случайного значения из списка.
        :param kwargs:
            choices - список значений
        :return: случайное значение
        """
        choices = kwargs.get('choices', [])
        if not choices:
            raise ValueError('Key choices not found!')

        return str(random.choice(seq=choices))  # nosec

    def mutation_numeric_smallint(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата smallint.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs.get('start', self.min_value_smallint)
        end = kwargs.get('end', self.max_value_smallint)
        if start < self.min_value_smallint or end > self.max_value_smallint:
            raise ValueError(
                f'The start and end values must be between {self.min_value_smallint} and {self.max_value_smallint}',
            )

        if kwargs.get('unique'):
            return str(self._generate_unique_value(func=self._numeric.integer_number, start=start, end=end))

        return str(self._numeric.integer_number(start=start, end=end))

    def mutation_numeric_integer(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата int.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs.get('start', self.min_value_integer)
        end = kwargs.get('end', self.max_value_integer)
        if start < self.min_value_integer or end > self.max_value_integer:
            raise ValueError(
                f'The start and end values must be between {self.min_value_integer} and {self.max_value_integer}',
            )

        if kwargs.get('unique'):
            return str(self._generate_unique_value(func=self._numeric.integer_number, start=start, end=end))

        return str(self._numeric.integer_number(start=start, end=end))

    def mutation_numeric_bigint(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата bigint.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs.get('start', self.min_value_bigint)
        end = kwargs.get('end', self.max_value_bigint)
        if start < self.min_value_bigint or end > self.max_value_bigint:
            raise ValueError(
                f'The start and end values must be between {self.min_value_bigint} and {self.max_value_bigint}',
            )

        if kwargs.get('unique'):
            return str(self._generate_unique_value(func=self._numeric.integer_number, start=start, end=end))

        return str(self._numeric.integer_number(start=start, end=end))

    def mutation_numeric_decimal(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата decimal.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            precision - количество символов после запятой
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs['start']
        end = kwargs['end']
        precision = kwargs['precision']
        if kwargs.get('unique'):
            return str(
                self._generate_unique_value(
                    func=self._numeric.float_number,
                    start=start,
                    end=end,
                    precision=precision,
                ),
            )

        return str(self._numeric.float_number(start=start, end=end, precision=precision))

    def mutation_numeric_real(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата real.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs['start']
        end = kwargs['end']
        if kwargs.get('unique'):
            return str(
                self._generate_unique_value(
                    func=self._numeric.float_number,
                    start=start,
                    end=end,
                    precision=6,
                ),
            )

        return str(self._numeric.float_number(start=start, end=end, precision=6))

    def mutation_numeric_double_precision(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата double precision.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs['start']
        end = kwargs['end']
        if kwargs.get('unique'):
            return str(
                self._generate_unique_value(
                    func=self._numeric.float_number,
                    start=start,
                    end=end,
                    precision=15,
                ),
            )

        return str(self._numeric.float_number(start=start, end=end, precision=15))

    def mutation_numeric_smallserial(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата smallserial.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs.get('start', self.min_value_smallserial)
        end = kwargs.get('end', self.max_value_smallserial)
        if start < self.min_value_smallserial or end > self.max_value_smallserial:
            raise ValueError(
                f'The start and end values must be between {self.min_value_smallserial} '
                f'and {self.max_value_smallserial}',
            )

        if kwargs.get('unique'):
            return str(self._generate_unique_value(func=self._numeric.integer_number, start=start, end=end))

        return str(self._numeric.integer_number(start=start, end=end))

    def mutation_numeric_serial(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата serial.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs.get('start', self.min_value_serial)
        end = kwargs.get('end', self.max_value_serial)
        if start < self.min_value_serial or end > self.max_value_serial:
            raise ValueError(
                f'The start and end values must be between {self.min_value_serial} and {self.max_value_serial}',
            )

        if kwargs.get('unique'):
            return str(self._generate_unique_value(func=self._numeric.integer_number, start=start, end=end))

        return str(self._numeric.integer_number(start=start, end=end))

    def mutation_numeric_bigserial(self, **kwargs: int) -> str:
        """
        Метод для формирования случайного числа формата bigserial.
        :param kwargs:
            start - минимальное значение
            end - максимальное значение
            unique - сгенерировать уникальное значение
        :return: случайное значение в пределах [start, end]
        """
        start = kwargs.get('start', self.min_value_bigserial)
        end = kwargs.get('end', self.max_value_bigserial)
        if start < self.min_value_bigserial or end > self.max_value_bigserial:
            raise ValueError(
                f'The start and end values must be between {self.min_value_bigserial} and {self.max_value_bigserial}',
            )

        if kwargs.get('unique'):
            return str(self._generate_unique_value(func=self._numeric.integer_number, start=start, end=end))

        return str(self._numeric.integer_number(start=start, end=end))

    def mutation_string_by_mask(self, **kwargs: Any) -> str:
        """
        Метод для формирования строки по маске.
        :param kwargs:
            mask - маска
            char - маска для символов
            digit - маска для цифр
            unique - сгенерировать уникальное значение
        """
        mask = kwargs['mask']
        char = kwargs.get('char', '@')
        digit = kwargs.get('digit', '#')
        if kwargs.get('unique'):
            return self._generate_unique_value(func=self._generate_string_by_mask, mask=mask, char=char, digit=digit)

        return self._generate_string_by_mask(mask=mask, char=char, digit=digit)

    @staticmethod
    def mutation_uuid4(**kwargs: Any) -> str:
        """
        Метод для формирования uuid4
        :param kwargs: параметры мутации - не используются
        :return: строка uuid4
        """
        return str(uuid.uuid4())

    def mutation_uuid5_by_source_value(self, **kwargs: Any) -> str:
        """
        Метод для формирования uuid5 с использованием значения из указанной колонки,
        текущей даты и переданного uuid namespace.
        :param kwargs:
            source_column: название колонки, значение которой хотим использовать.
            namespace: uuid namespace.
        :return: строка uuid5.
        """
        source_column: Optional[str] = kwargs.get('source_column')
        if not source_column:
            raise ValueError('Argument "source_column" not found')

        obfuscated_values: dict[str, Any] = kwargs.get('obfuscated_values', {})
        source_value: Optional[str] = obfuscated_values.get(source_column)
        if not source_value:
            raise ValueError('Value of "source_column" not found')

        namespace: Optional[str] = kwargs.get('namespace')
        if namespace is None:
            raise ValueError('Argument "namespace" not found')

        try:
            uuid_namespace: uuid.UUID = uuid.UUID(namespace)
        except Exception as e:
            raise ValueError('Invalid uuid namespace given') from e

        return str(uuid.uuid5(uuid_namespace, f'{source_value}-{self._today}'))



from setuptools import setup
from os.path import join, dirname

setup(
    name='pg_stage',
    version='0.3.5',
    packages=['pg_stage'],
    package_dir={'': 'src'},
    long_description=open(join(dirname(__file__), 'README.md')).read(),
    install_requires=['typing-extensions>=4.5.0', 'mimesis==4.1.3'],
    extras_require={'dev': ['pytest']},
    include_package_data=True,
    license_files=('LICENSE.txt',),
)



import pytest

from src.pg_stage.obfuscator import Obfuscator


@pytest.mark.parametrize("table_name", ['table', 'schema.table'])
def test_parse_comment_table_with_mutation(obfuscator_object: Obfuscator, table_name: str):
    """
    Arrange: Строка, в которой есть комментарий к таблице
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Таблица добавлена в список _delete_tables
    """
    obfuscator_object._parse_line(
        line=f'COMMENT ON TABLE {table_name}' + ' IS \'anon: {"mutation_name": "delete"}\';',
    )
    assert table_name in obfuscator_object._delete_tables  # nosec


@pytest.mark.parametrize("table_name", ['table', 'schema.table'])
def test_parse_comment_table_with_invalid_mutation(obfuscator_object: Obfuscator, table_name: str):
    """
    Arrange: Строка, в которой есть невалидный комментарий к таблице
    Act: Вызов функции `_parse_line` класса Obfuscator
    Assert: Мутация не добавлена в список _delete_tables
    """
    obfuscator_object._parse_line(
        line=f'COMMENT ON TABLE {table_name}' + ' IS \'anon: {"mutation_name": "set"}\';',
    )
    assert table_name not in obfuscator_object._delete_tables  # nosec



from typing import List, Callable, Any, Dict
from enum import Enum

from typing_extensions import TypedDict


class OperationChoices(Enum):
    """Описание видов операций"""

    equal = 'equal'
    not_equal = 'not_equal'
    by_pattern = 'by_pattern'


class ConditionType(TypedDict):
    """Описание типа условий для обфускации"""

    column_name: str
    operation: OperationChoices
    value: str


class RelationType(TypedDict):
    """Описание типа зависимых таблиц"""

    table_name: str
    column_name: str
    from_column_name: str
    to_column_name: str


ConditionTypeMany = List[ConditionType]
RelationTypeMany = List[RelationType]


class MapTablesValueType(TypedDict):
    """Описание типа значения карты таблиц"""

    mutation_name: str
    mutation_func: Callable[..., str]
    mutation_kwargs: Dict[str, Any]
    mutation_relations: RelationTypeMany
    mutation_conditions: ConditionTypeMany


MapTablesValueTypeMany = List[MapTablesValueType]
